"""
Module for printing & visualizing Jikoku object data.
"""
import itertools
import logging

from jikoku.models import *
from jikoku.scheduler import schedule
from tests.test_scheduler_basic import get_basic_schedule

Log = logging.getLogger(__name__)


def pretty_print_schedule(schedule: Schedule) -> str:
    to_print = f"\nPrinting Schedule'{schedule.name}'\n"
    # sort schedule by train:
    sorted_trips = sorted(schedule.trips, key=lambda t: t.train.name)

    # group schedule by train
    for train, trips in itertools.groupby(sorted_trips, key=lambda t: t.train.name):
        to_print += f"{train}\n"
        for trip in trips:
            to_print += f"\t{trip.service.start_time} - {trip.service.end_time}: {trip.service.first_stop.name} => {trip.service.last_stop.name}\n"

    return to_print


if __name__ == "__main__":
    loglevel = logging.DEBUG
    logging.basicConfig(
        level=loglevel,
        format="%(asctime)s - %(levelname)s@%(module)s:%(lineno)d - %(message)s",
        datefmt="[%H:%M:%S]",
    )

    s = schedule(get_basic_schedule())
    Log.info(pretty_print_schedule(s))



from dataclasses import dataclass
from datetime import time, timedelta
from typing import Self

from jikoku.time import TimePoint
from jikoku.utils import add_times, subtract_times


@dataclass
class Stop:
    name: str
    stop_time: TimePoint

    def __add__(self, other):
        return Stop(self.name, self.stop_time.plus(other))

    def __sub__(self, other):
        return Stop(self.name, self.stop_time.minus(other))

    def __le__(self, other):
        pass


@dataclass
class Service:
    """
    This is a test api documentation
    """

    name: str
    stops: list[Stop]

    @classmethod
    def from_stops(cls, name: str, stops: list[Stop]) -> Self:
        return cls(name, stops)

    @property
    def start_time(self) -> TimePoint:
        return self.stops[0].stop_time

    @property
    def first_stop(self) -> Stop:
        return self.stops[0]

    @property
    def last_stop(self) -> Stop:
        return self.stops[-1]

    @property
    def end_time(self) -> TimePoint:
        return self.stops[-1].stop_time

    def __add__(self, other: TimePoint):
        new_name = self.name + "_shift_by_" + str(other)
        return Service(
            new_name,
            [stop + other for stop in self.stops],
        )

    def __sub__(self, other: TimePoint):
        new_name = self.name + "_shift_by_minus" + str(other)
        return Service(
            new_name,
            [stop - other for stop in self.stops],
        )


@dataclass
class Train:
    # TODO this should be unique, any way to enforce?
    name: str


@dataclass
class Trip:
    service: Service
    train: Train


@dataclass
class Schedule:
    name: str
    trips: list[Trip]

    @property
    def number_of_trains(self):
        return len({t.train.name for t in self.trips})



import logging
import os
import secrets
import string
from datetime import time, date, timedelta, datetime


def add_times(a_time: time, a_time_delta: timedelta):
    """
    Adds a timedelta to a time, ignoring 24h rollover.
    """
    return (datetime.combine(date.today(), a_time) + a_time_delta).time()


def subtract_times(a_time: time, a_time_delta: timedelta):
    """
    Adds a timedelta to a time, ignoring 24h rollover.
    """
    return (datetime.combine(date.today(), a_time) - a_time_delta).time()


def compare_times(a_time: time, b_time: time) -> timedelta:
    """
    Compares two times, ignoring 24 rollover, i.e. 23h - 1h the next day will give a time difference of 22 hours!
    """
    return datetime.combine(date.today(), a_time) - datetime.combine(
        date.today(), b_time
    )


def generate_unique_name(length: int = 6) -> str:
    characters = string.ascii_letters + string.digits
    unique_string = "".join(secrets.choice(characters) for _ in range(length))
    return unique_string


def setup_logging():
    loglevel = logging.DEBUG if os.getenv("DEBUG") else logging.WARNING
    logging.basicConfig(
        level=loglevel,
        format="%(asctime)s - %(levelname)s@%(module)s:%(lineno)d - %(message)s",
        datefmt="[%H:%M:%S]",
    )



from abc import ABC, abstractmethod
from typing import Self

# FIXME: the abstract class should have some notion of adding a few minutes to an existing timepoint!


class TimePoint(ABC):
    @abstractmethod
    def plus(self, other: Self) -> Self:
        pass

    @abstractmethod
    def minus(self, other: Self) -> Self:
        pass

    @abstractmethod
    def same_time(self, other: Self) -> bool:
        pass

    @abstractmethod
    def earlier_than(self, other: Self) -> bool:
        pass

    def __add__(self, other):
        if not isinstance(other, self.__class__):
            raise TypeError(
                "Can only add a TimePoint of the same type to another TimePoint!"
            )

        return self.plus(other)

    def __sub__(self, other):
        if not isinstance(other, self.__class__):
            raise TypeError(
                "Can only subtract a TimePoint of the same type to another TimePoint!"
            )

        return self.minus(other)

    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            raise TypeError("Can only compare Timepoints of the same type!")

        return self.same_time(other)

    def __lt__(self, other):
        if not isinstance(other, self.__class__):
            raise TypeError("Can only compare Timepoints of the same type!")

        return self.same_time(other)

    @abstractmethod
    def __str__(self):
        """Forces child classes to implement __str__ method"""
        pass


class DailyTimePoint(TimePoint, ABC):
    def __init__(
        self, hour: int = 0, minute: int = 0, day: int = 0, previous: Self = None
    ):
        self._t = day * 24 * 60 + hour * 60 + minute

    def plus(self, other: Self) -> Self:
        return DailyTimePoint(hour=0, minute=self._t + other._t)

    def minus(self, other: Self) -> Self:
        return DailyTimePoint(hour=0, minute=self._t - other._t)

    def same_time(self, other: Self) -> bool:
        return self._t == other._t

    def earlier_than(self, other: Self) -> bool:
        return self._t <= other._t

    def __str__(self):
        date, total_min = divmod(self._t, 24 * 60)
        hours, minutes = divmod(total_min, 60)

        if date == 0:
            return f"{hours:02}:{minutes:02}"
        elif date == 1:
            return f"Tomorrow at {hours:02}:{minutes:02}"
        elif date == -1:
            return f"Yesterday at {hours:02}:{minutes:02}"
        elif date >= 2:
            return f"In {date} days at {hours:02}:{minutes:02}"
        elif date <= -2:
            return f"{abs(date)} days ago at {hours:02}:{minutes:02}"


class WeeklyTimePoint(TimePoint):
    pass



import logging
import os
from dataclasses import dataclass

import jikoku.config as config
from jikoku.models import *
from jikoku.time import DailyTimePoint
from jikoku.utils import *

Log = logging.getLogger(__name__)


@dataclass
class TrainData:
    train: Train
    busy_until: TimePoint
    free_location: str

    def available(self, at_time: TimePoint) -> bool:
        return self.busy_until.earlier_than(at_time)


def choose_train_heuristic(to_fill: Service, available_trains: list[TrainData]):
    # give precedence to trains at the same location
    trains_at_same_location = [
        t for t in available_trains if t.free_location == to_fill.first_stop.name
    ]
    if trains_at_same_location:
        return trains_at_same_location[0]
    else:
        return available_trains[0]


def schedule(services: list[Service], **kwargs) -> Schedule:
    conf = config.Config()
    for kword, value in kwargs.items():
        if kword in conf.__dict__:
            conf.__dict__[kword] = value
        else:
            Log.error(f"unknown config item {kword}, skipping")

    Log.info(f"using config {config}")

    services.sort(key=lambda s: s.start_time)

    trainsData: list[TrainData] = list()
    trips = list()

    for service in services:
        current_time = service.start_time

        Log.debug(f"Scheduling service {service}")

        Log.debug(f"Current traindata status {service}")

        available_trains = [t for t in trainsData if t.available(current_time)]

        if not available_trains:
            Log.debug("No train available, making a new one")
            using_train = Train(f"train-{generate_unique_name()}")
            data = TrainData(
                using_train,
                (
                    service.end_time
                    + DailyTimePoint(minute=conf.minimum_minutes_between_services)
                ),
                service.last_stop.name,
            )
            trainsData.append(data)
        else:
            data = choose_train_heuristic(service, available_trains)
            using_train = data.train
            data.busy_until = service.end_time + DailyTimePoint(
                minute=conf.minimum_minutes_between_services
            )
            data.free_location = service.last_stop.name

        trips.append(Trip(service, using_train))

    return Schedule("autogenerated schedule", trips)
